var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
import * as React from "react";
import { FLOW_UPDATE_EVENT, getKey, makeKey, } from "../lib";
const defaultFlowContext = {
    get controller() {
        throw new Error("No flow provider found");
    },
};
const flowContext = React.createContext(defaultFlowContext);
export function useController() {
    return React.useContext(flowContext).controller;
}
export function usePrefetcher() {
    const controller = useController();
    const keys = React.useRef(new Set()).current;
    const rerender = React.useState()[1];
    React.useEffect(() => {
        return controller.on(FLOW_UPDATE_EVENT, (flow) => {
            if (!keys.has(flow.key))
                return;
            rerender({});
        });
    }, [keys, rerender, controller]);
    return React.useMemo(() => [
        (...args) => {
            const key = Array.isArray(args[0]) ? makeKey(args[0]) : args[0];
            keys.add(key);
            const flow = controller.start(...args);
            return flow;
        },
        (key, ...args) => {
            const [fn, mapper, defaultValue] = typeof args[0] === "function" && typeof args[1] === "function"
                ? args
                : [undefined, ...args];
            const k = Array.isArray(key) ? makeKey(key) : key;
            const flow = fn ? controller.flow(k, fn) : controller.flow(k);
            return flow ? mapper(flow) : defaultValue;
        },
    ], [controller, keys]);
}
export function useFlow(...args) {
    let fixedArgs = false;
    const originalKey = args[0];
    // key array  useFlow([key, arg1, arg2, arg3, arg4])
    if (Array.isArray(args[0])) {
        // overwrite args option
        // exclude first item (key)
        args[2] = Object.assign(Object.assign({}, args[2]), { args: args[0].slice(1) });
        args[0] = makeKey(args[0]);
        fixedArgs = true;
    }
    const provider = React.useContext(flowContext);
    let useDefaultFlow = false;
    let overrideArgs;
    if (typeof args[1] === "function") {
        overrideArgs = args;
    }
    // useFlow(fn, options)
    else if (typeof args[0] === "function") {
        overrideArgs = [getKey(args[0]), args[0], args[1]];
    }
    else {
        // useFlow(key, options)
        useDefaultFlow = true;
        overrideArgs = [getKey(args[0]), provider.defaultFlow, args[1]];
    }
    const [key, fn, _a] = overrideArgs, _b = _a === void 0 ? {} : _a, { suspense = provider.suspense, errorBoundary = provider.errorBoundary } = _b, options = __rest(_b, ["suspense", "errorBoundary"]);
    const rerender = React.useState()[1];
    const optionsRef = React.useRef();
    const renderingRef = React.useRef(false);
    const flowRef = React.useRef();
    const { flowHook, handleSuspeseAndErrorBoundary } = React.useMemo(() => {
        return createFlowHook(flowRef, renderingRef, optionsRef, fixedArgs);
    }, [suspense, errorBoundary, fixedArgs]);
    // update refs
    renderingRef.current = true;
    optionsRef.current = Object.assign(Object.assign({}, options), { suspense,
        errorBoundary, prependArgs: useDefaultFlow
            ? Array.isArray(originalKey)
                ? originalKey
                : [originalKey]
            : [] });
    // make sure removed flow does not cause an error
    flowRef.current =
        provider.controller.flow(key, fn) || flowRef.current;
    React.useLayoutEffect(() => {
        renderingRef.current = false;
        return provider.controller.on(FLOW_UPDATE_EVENT, (flow) => {
            if (flow.key === key) {
                rerender({});
            }
        });
    });
    handleSuspeseAndErrorBoundary();
    return flowHook;
}
function createFlowHook(flowRef, renderingRef, optionsRef, fixedArgs) {
    function handleSuspeseAndErrorBoundary() {
        var _a, _b;
        if (!flowRef.current || !renderingRef.current)
            return;
        if (((_a = optionsRef.current) === null || _a === void 0 ? void 0 : _a.errorBoundary) &&
            flowRef.current.status === "faulted") {
            throw flowRef.current.error;
        }
        if (((_b = optionsRef.current) === null || _b === void 0 ? void 0 : _b.suspense) && flowRef.current.status === "running") {
            throw new Promise((resolve) => {
                var _a;
                (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.on("update", resolve);
            });
        }
    }
    function getArgs(inputArgs) {
        var _a, _b;
        const prependArgs = ((_a = optionsRef.current) === null || _a === void 0 ? void 0 : _a.prependArgs) || [];
        const defaultArgs = (_b = optionsRef.current) === null || _b === void 0 ? void 0 : _b.args;
        if (!defaultArgs) {
            return [...prependArgs, ...inputArgs];
        }
        return [
            ...prependArgs,
            ...inputArgs,
            ...defaultArgs.slice(inputArgs.length),
        ];
    }
    function run(type, args) {
        var _a;
        if (args.length && fixedArgs) {
            throw new Error("Passing arguments to fixed args flow is not allowed. Use the overload useFlow(key, flow, { args: [] }) instead");
        }
        (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a[type](...getArgs(args));
        handleSuspeseAndErrorBoundary();
        return flowHook;
    }
    const flowHook = {
        get hasData() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.hasData) || false;
        },
        get status() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.status) || "unknown";
        },
        get error() {
            var _a;
            return (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.error;
        },
        get running() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.running) || false;
        },
        get completed() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.completed) || false;
        },
        get faulted() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.faulted) || false;
        },
        get idle() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.idle) || false;
        },
        get data() {
            var _a;
            return (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.data;
        },
        get current() {
            var _a;
            return (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.current;
        },
        get cancelled() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.cancelled) || false;
        },
        get stale() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.stale) || false;
        },
        update(data) {
            var _a;
            (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.update(data);
            return flowHook;
        },
        start(...args) {
            return run("start", args);
        },
        restart(...args) {
            return run("restart", args);
        },
        next(payload) {
            var _a;
            (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.next(payload);
            return flowHook;
        },
        cancel() {
            var _a;
            (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.cancel();
            return flowHook;
        },
    };
    return { flowHook, handleSuspeseAndErrorBoundary };
}
export const FlowProvider = (props) => {
    const parentProvider = React.useContext(flowContext);
    const { controller = parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.controller, suspense = (parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.suspense) || false, errorBoundary = (parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.errorBoundary) || false, defaultFlow = parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.defaultFlow, children, } = props;
    const value = React.useMemo(() => ({ controller, suspense, errorBoundary, defaultFlow }), [controller, suspense, errorBoundary, defaultFlow]);
    if (!controller) {
        throw new Error("No controller provided");
    }
    if (controller.error) {
        throw controller.error;
    }
    if (!controller.ready) {
        if (suspense)
            throw controller.promise;
        return null;
    }
    return React.createElement(flowContext.Provider, {
        value,
        children,
    });
};
