import * as React from "react";
import { AnyFunc, FlowController, FlowDataInfer, FlowStatus, Flow, FlowPrefetcher } from "../lib";
export interface FlowProviderProps {
    controller?: FlowController;
    defaultFlow?: AnyFunc;
    suspense?: boolean;
    errorBoundary?: boolean;
}
export interface FlowHookWithoutArgs<T extends AnyFunc> {
    readonly current: Flow<T> | undefined;
    readonly data: FlowDataInfer<T> | undefined;
    readonly status: FlowStatus | "unknown";
    readonly idle: boolean;
    readonly running: boolean;
    readonly completed: boolean;
    readonly faulted: boolean;
    readonly error: Error | undefined;
    readonly cancelled: boolean;
    readonly stale: boolean;
    readonly hasData: boolean;
    update(data: FlowDataInfer<T>): this;
    update(reducer: (prev: FlowDataInfer<T>) => FlowDataInfer<T>): this;
    next(payload?: any): this;
    start(): this;
    restart(): this;
    cancel(): this;
}
export interface FlowHook<T extends AnyFunc> extends FlowHookWithoutArgs<T> {
    start(...args: Parameters<T>): this;
    restart(...args: Parameters<T>): this;
}
export interface UseFlowOptions<T extends AnyFunc> extends UseFlowOptionsWithoutArgs {
    args?: Parameters<T>;
    suspense?: boolean;
    errorBoundary?: boolean;
}
export interface UseFlowOptionsWithoutArgs {
}
export interface PrefetchMapFn extends Function {
    <T extends AnyFunc, TResult>(key: any, flow: T, mapper: (flow: Flow) => TResult, defaultValue?: TResult): TResult | undefined;
    <TResult>(key: any, mapper: (flow: Flow) => TResult, defaultValue?: TResult): TResult | undefined;
}
export declare function useController(): FlowController;
export declare function usePrefetcher(): [FlowPrefetcher, PrefetchMapFn];
export declare function useFlow<T extends AnyFunc>(key: [string, ...Parameters<T>], flow: T, options?: UseFlowOptionsWithoutArgs): FlowHookWithoutArgs<T>;
export declare function useFlow<T extends AnyFunc>(key: string, flow: T, options?: UseFlowOptions<T>): FlowHook<T>;
export declare function useFlow<T extends AnyFunc>(flow: T, options?: UseFlowOptions<T>): FlowHook<T>;
export declare function useFlow(key: string | [string, ...any[]], options?: UseFlowOptionsWithoutArgs): FlowHook<AnyFunc>;
export declare const FlowProvider: React.FC<FlowProviderProps>;
