"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.FlowProvider = exports.useFlow = exports.usePrefetcher = exports.useController = void 0;
var React = require("react");
var lib_1 = require("../lib");
var defaultFlowContext = {
    get controller() {
        throw new Error("No flow provider found");
    },
};
var flowContext = React.createContext(defaultFlowContext);
function useController() {
    return React.useContext(flowContext).controller;
}
exports.useController = useController;
function usePrefetcher() {
    var controller = useController();
    var keys = React.useRef(new Set()).current;
    var rerender = React.useState()[1];
    React.useEffect(function () {
        return controller.on(lib_1.FLOW_UPDATE_EVENT, function (flow) {
            if (!keys.has(flow.key))
                return;
            rerender({});
        });
    }, [keys, rerender, controller]);
    return React.useMemo(function () { return [
        function () {
            var _a;
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            var key = Array.isArray(args[0]) ? (0, lib_1.makeKey)(args[0]) : args[0];
            keys.add(key);
            var flow = (_a = controller).start.apply(_a, args);
            return flow;
        },
        function (key) {
            var args = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                args[_i - 1] = arguments[_i];
            }
            var _a = typeof args[0] === "function" && typeof args[1] === "function"
                ? args
                : __spreadArray([undefined], args, true), fn = _a[0], mapper = _a[1], defaultValue = _a[2];
            var k = Array.isArray(key) ? (0, lib_1.makeKey)(key) : key;
            var flow = fn ? controller.flow(k, fn) : controller.flow(k);
            return flow ? mapper(flow) : defaultValue;
        },
    ]; }, [controller, keys]);
}
exports.usePrefetcher = usePrefetcher;
function useFlow() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    var fixedArgs = false;
    var originalKey = args[0];
    // key array  useFlow([key, arg1, arg2, arg3, arg4])
    if (Array.isArray(args[0])) {
        // overwrite args option
        // exclude first item (key)
        args[2] = __assign(__assign({}, args[2]), { args: args[0].slice(1) });
        args[0] = (0, lib_1.makeKey)(args[0]);
        fixedArgs = true;
    }
    var provider = React.useContext(flowContext);
    var useDefaultFlow = false;
    var overrideArgs;
    if (typeof args[1] === "function") {
        overrideArgs = args;
    }
    // useFlow(fn, options)
    else if (typeof args[0] === "function") {
        overrideArgs = [(0, lib_1.getKey)(args[0]), args[0], args[1]];
    }
    else {
        // useFlow(key, options)
        useDefaultFlow = true;
        overrideArgs = [(0, lib_1.getKey)(args[0]), provider.defaultFlow, args[1]];
    }
    var key = overrideArgs[0], fn = overrideArgs[1], _a = overrideArgs[2], _b = _a === void 0 ? {} : _a, _c = _b.suspense, suspense = _c === void 0 ? provider.suspense : _c, _d = _b.errorBoundary, errorBoundary = _d === void 0 ? provider.errorBoundary : _d, options = __rest(_b, ["suspense", "errorBoundary"]);
    var rerender = React.useState()[1];
    var optionsRef = React.useRef();
    var renderingRef = React.useRef(false);
    var flowRef = React.useRef();
    var _e = React.useMemo(function () {
        return createFlowHook(flowRef, renderingRef, optionsRef, fixedArgs);
    }, [suspense, errorBoundary, fixedArgs]), flowHook = _e.flowHook, handleSuspeseAndErrorBoundary = _e.handleSuspeseAndErrorBoundary;
    // update refs
    renderingRef.current = true;
    optionsRef.current = __assign(__assign({}, options), { suspense: suspense, errorBoundary: errorBoundary, prependArgs: useDefaultFlow
            ? Array.isArray(originalKey)
                ? originalKey
                : [originalKey]
            : [] });
    // make sure removed flow does not cause an error
    flowRef.current =
        provider.controller.flow(key, fn) || flowRef.current;
    React.useLayoutEffect(function () {
        renderingRef.current = false;
        return provider.controller.on(lib_1.FLOW_UPDATE_EVENT, function (flow) {
            if (flow.key === key) {
                rerender({});
            }
        });
    });
    handleSuspeseAndErrorBoundary();
    return flowHook;
}
exports.useFlow = useFlow;
function createFlowHook(flowRef, renderingRef, optionsRef, fixedArgs) {
    function handleSuspeseAndErrorBoundary() {
        var _a, _b;
        if (!flowRef.current || !renderingRef.current)
            return;
        if (((_a = optionsRef.current) === null || _a === void 0 ? void 0 : _a.errorBoundary) &&
            flowRef.current.status === "faulted") {
            throw flowRef.current.error;
        }
        if (((_b = optionsRef.current) === null || _b === void 0 ? void 0 : _b.suspense) && flowRef.current.status === "running") {
            throw new Promise(function (resolve) {
                var _a;
                (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.on("update", resolve);
            });
        }
    }
    function getArgs(inputArgs) {
        var _a, _b;
        var prependArgs = ((_a = optionsRef.current) === null || _a === void 0 ? void 0 : _a.prependArgs) || [];
        var defaultArgs = (_b = optionsRef.current) === null || _b === void 0 ? void 0 : _b.args;
        if (!defaultArgs) {
            return __spreadArray(__spreadArray([], prependArgs, true), inputArgs, true);
        }
        return __spreadArray(__spreadArray(__spreadArray([], prependArgs, true), inputArgs, true), defaultArgs.slice(inputArgs.length), true);
    }
    function run(type, args) {
        var _a;
        if (args.length && fixedArgs) {
            throw new Error("Passing arguments to fixed args flow is not allowed. Use the overload useFlow(key, flow, { args: [] }) instead");
        }
        (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a[type].apply(_a, getArgs(args));
        handleSuspeseAndErrorBoundary();
        return flowHook;
    }
    var flowHook = {
        get hasData() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.hasData) || false;
        },
        get status() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.status) || "unknown";
        },
        get error() {
            var _a;
            return (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.error;
        },
        get running() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.running) || false;
        },
        get completed() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.completed) || false;
        },
        get faulted() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.faulted) || false;
        },
        get idle() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.idle) || false;
        },
        get data() {
            var _a;
            return (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.data;
        },
        get current() {
            var _a;
            return (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.current;
        },
        get cancelled() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.cancelled) || false;
        },
        get stale() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.stale) || false;
        },
        update: function (data) {
            var _a;
            (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.update(data);
            return flowHook;
        },
        start: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return run("start", args);
        },
        restart: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return run("restart", args);
        },
        next: function (payload) {
            var _a;
            (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.next(payload);
            return flowHook;
        },
        cancel: function () {
            var _a;
            (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.cancel();
            return flowHook;
        },
    };
    return { flowHook: flowHook, handleSuspeseAndErrorBoundary: handleSuspeseAndErrorBoundary };
}
var FlowProvider = function (props) {
    var parentProvider = React.useContext(flowContext);
    var _a = props.controller, controller = _a === void 0 ? parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.controller : _a, _b = props.suspense, suspense = _b === void 0 ? (parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.suspense) || false : _b, _c = props.errorBoundary, errorBoundary = _c === void 0 ? (parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.errorBoundary) || false : _c, _d = props.defaultFlow, defaultFlow = _d === void 0 ? parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.defaultFlow : _d, children = props.children;
    var value = React.useMemo(function () { return ({ controller: controller, suspense: suspense, errorBoundary: errorBoundary, defaultFlow: defaultFlow }); }, [controller, suspense, errorBoundary, defaultFlow]);
    if (!controller) {
        throw new Error("No controller provided");
    }
    if (controller.error) {
        throw controller.error;
    }
    if (!controller.ready) {
        if (suspense)
            throw controller.promise;
        return null;
    }
    return React.createElement(flowContext.Provider, {
        value: value,
        children: children,
    });
};
exports.FlowProvider = FlowProvider;
